from dataclasses import dataclass
from functools import cache, reduce
from heapq import heapify, heappop, heappush
from typing import Literal

import numpy as np
import numpy.typing as npt
from tqdm import tqdm
from typing_extensions import Self

from .._util.docs import mkdocstrings
from .._util.validators import validate_pmf
from ..types import Array1D
from .FixedToVariableCode import FixedToVariableCode
from .util import Word


@mkdocstrings(filters=["!.*"])
class HuffmanCode(FixedToVariableCode):
    r"""
    Binary Huffman code. It is an optimal (minimal expected rate) [fixed-to-variable length code](/ref/FixedToVariableCode) for a given pmf $p$ over $\mathcal{X}$. For more details, see <cite>Say06, Sec. 3.2</cite>.

    Notes:
        Huffman codes are always [prefix-free](/ref/FixedToVariableCode/#is_prefix_free) (hence [uniquely decodable](/ref/FixedToVariableCode/#is_uniquely_decodable)).

    Parameters:
        pmf: The pmf $p$ to be considered. It must be a one-dimensional array of floats of size $|\mathcal{X}|$. The elements must be non-negative and sum to $1$.

        source_block_size: The source block size $k$. The default value is $k = 1$.

        policy: The policy to be used when constructing the code. It must be either `'high'` (move combined symbols as high as possible) or `'low'` (move combined symbols as low as possible). The default value is `'high'`.

    Examples:
        >>> pmf = [0.8, 0.1, 0.1]

        >>> code = komm.HuffmanCode(pmf)
        >>> code.enc_mapping
        {(0,): (0,),
         (1,): (1, 0),
         (2,): (1, 1)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(1.2)

        >>> code = komm.HuffmanCode(pmf, 2)
        >>> code.enc_mapping
        {(0, 0): (0,),
         (0, 1): (1, 0, 1),
         (0, 2): (1, 1, 0),
         (1, 0): (1, 1, 1),
         (1, 1): (1, 0, 0, 1, 0, 0),
         (1, 2): (1, 0, 0, 1, 0, 1),
         (2, 0): (1, 0, 0, 0),
         (2, 1): (1, 0, 0, 1, 1, 0),
         (2, 2): (1, 0, 0, 1, 1, 1)}
        >>> code.rate(pmf)  # doctest: +FLOAT_CMP
        np.float64(0.96)
    """

    def __init__(
        self,
        pmf: npt.ArrayLike,
        source_block_size: int = 1,
        policy: Literal["high", "low"] = "high",
    ):
        self.pmf = validate_pmf(pmf)
        if not source_block_size >= 1:
            raise ValueError("'source_block_size' must be at least 1")
        if not policy in {"high", "low"}:
            raise ValueError("'policy' must be in {'high', 'low'}")
        self.policy = policy
        super().__init__(
            source_cardinality=self.pmf.size,
            target_cardinality=2,
            source_block_size=source_block_size,
            enc_mapping=huffman_code(self.pmf, source_block_size, policy),
        )

    def __repr__(self) -> str:
        args = ", ".join([
            f"pmf={self.pmf.tolist()}",
            f"source_block_size={self.source_block_size}",
            f"policy={self.policy!r}",
        ])
        return f"{self.__class__.__name__}({args})"

    @cache
    def is_uniquely_decodable(self) -> bool:
        return True

    @cache
    def is_prefix_free(self) -> bool:
        return True


def huffman_code_lengths(pmf: Array1D[np.floating]) -> Array1D[np.integer]:
    lengths = np.zeros_like(pmf, dtype=int)
    if np.sum(pmf**2) == 1:  # Deterministic case
        lengths[pmf > 0] = 1
        return lengths
    heap: list[tuple[np.floating, list[int]]] = []
    for i, pi in enumerate(pmf):
        if pi > 0:
            heappush(heap, (pi, [i]))
    while len(heap) > 1:
        w1, s1 = heappop(heap)
        w2, s2 = heappop(heap)
        for i in s1:
            lengths[i] += 1
        for i in s2:
            lengths[i] += 1
        heappush(heap, (w1 + w2, s1 + s2))
    return lengths


def huffman_code(
    pmf: Array1D[np.floating],
    source_block_size: int,
    policy: Literal["high", "low"],
) -> dict[Word, Word]:

    @dataclass(slots=True)
    class Node:
        index: int
        probability: np.floating
        leaf: bool = True
        parent: int = -1
        bit: int = -1

        def __lt__(self, other: Self) -> bool:
            p0, i0 = self.probability, self.index
            p1, i1 = other.probability, other.index
            if policy == "high":
                return (p0, -i0 if self.leaf else i0) < (p1, -i1 if other.leaf else i1)
            elif policy == "low":
                return (p0, -i0) < (p1, -i1)

    extended_pmf = reduce(np.multiply.outer, [pmf] * source_block_size)

    pbar = tqdm(desc="Generating Huffman code", total=3 * extended_pmf.size, delay=2.5)
    pbar.update()

    tree: list[Node] = []
    for index, prob in enumerate(extended_pmf.ravel()):
        tree.append(Node(index, prob))
        pbar.update()

    heap = tree.copy()
    heapify(heap)
    while len(heap) > 1:
        node1 = heappop(heap)
        node0 = heappop(heap)
        node = Node(
            index=len(tree),
            probability=node0.probability + node1.probability,
            leaf=False,
        )
        node1.bit = 1
        node0.bit = 0
        node0.parent = node1.parent = node.index
        heappush(heap, node)
        tree.append(node)
        pbar.update()

    enc_mapping: dict[Word, Word] = {x: () for x in np.ndindex(extended_pmf.shape)}
    for index, x in enumerate(enc_mapping.keys()):
        node = tree[index]
        bits: list[int] = []
        while node.parent >= 0:
            bits.append(node.bit)
            node = tree[node.parent]
        y = tuple(reversed(bits))
        enc_mapping[x] = y
        pbar.update()

    pbar.close()

    return enc_mapping



from typing import Any

from tqdm import tqdm


def get_pbar(total: int, algorithm: str) -> "tqdm[Any]":
    return tqdm(
        total=total,
        desc=f"Decoding with {algorithm} algorithm",
        unit="block",
        delay=2.5,
    )



from typing import TypeVar

import numpy as np
import numpy.typing as npt

from ..types import Array1D

DType = TypeVar("DType", bound=np.floating | np.complexfloating)


def autocorrelation(
    sequence: npt.ArrayLike,
    shifts: npt.ArrayLike | None = None,
    normalized: bool = False,
) -> Array1D[DType]:
    r"""
    Computes the autocorrelation $R[\ell]$ of a real or complex sequence $x[n]$. This is defined as
    $$
        R[\ell] = \sum_{n \in \mathbb{Z}} x[n] x^*_\ell[n],
    $$
    where $x^\*\_\ell[n] = x^\*[n - \ell]$ is the complex conjugate of $x[n]$ shifted by $\ell$ positions. The autocorrelation $R[\ell]$ is even symmetric and satisfies $R[\ell] = 0$ for $|\ell| \geq L$, where $L$ is the length of the sequence.

    Parameters:
        sequence: A 1D-array containing the sequence $x[n]$, of length $L$.

        shifts: A 1D-array containing the values of $\ell$ for which the autocorrelation will be computed. The default value is `range(len(sequence))`, that is, $[0 : L)$.

        normalized: If `True`, returns the autocorrelation divided by the sequence energy, so that $R[0] = 1$. The default value is `False`.

    Returns:
        The autocorrelation $R[\ell]$ of the sequence.

    Examples:
        >>> komm.autocorrelation([1.0, 2.0, 3.0, 4.0], shifts=[-2, -1, 0, 1, 2])
        array([11., 20., 30., 20., 11.])
    """
    sequence = np.asarray(sequence)
    seq_conj = np.conj(sequence)
    shifts = np.arange(sequence.size) if shifts is None else np.asarray(shifts)
    acorr = np.empty_like(shifts, dtype=sequence.dtype)
    for i, ℓ in enumerate(shifts):
        if ℓ < 0:
            acorr[i] = np.dot(sequence[:ℓ], seq_conj[-ℓ:])
        elif ℓ > 0:
            acorr[i] = np.dot(sequence[ℓ:], seq_conj[:-ℓ])
        else:
            acorr[i] = np.dot(sequence, seq_conj)
    energy = np.dot(sequence, seq_conj)
    return acorr / energy if normalized else acorr


def cyclic_autocorrelation(
    sequence: npt.ArrayLike,
    shifts: npt.ArrayLike | None = None,
    normalized: bool = False,
) -> Array1D[DType]:
    r"""
    Computes the cyclic autocorrelation $\tilde{R}[\ell]$ of a real or complex sequence $x[n]$. This is defined as
    $$
        \tilde{R}[\ell] = \sum_{n \in [0:L)} x[n] \tilde{x}^*_\ell[n],
    $$
    where $\tilde{x}^\*\_\ell[n]$ is the complex conjugate of $x[n]$ cyclic-shifted by $\ell$ positions, and $L$ is the period of the sequence. The cyclic autocorrelation $\tilde{R}[\ell]$ is even symmetric and periodic with period $L$.

    Parameters:
        sequence: A 1D-array containing the sequence $x[n]$, of length $L$.

        shifts: A 1D-array containing the values of $\ell$ for which the cyclic autocorrelation will be computed. The default value is `range(len(sequence))`, that is, $[0 : L)$.

        normalized: If `True`, returns the cyclic autocorrelation divided by the sequence energy, so that $R[0] = 1$. The default value is `False`.

    Returns:
        The cyclic autocorrelation $\tilde{R}[\ell]$ of the sequence.

    Examples:
        >>> komm.cyclic_autocorrelation([1.0, 2.0, 3.0, 4.0], shifts=[-2, -1, 0, 1, 2])
        array([22., 24., 30., 24., 22.])
    """
    sequence = np.asarray(sequence)
    shifts = np.arange(sequence.size) if shifts is None else np.asarray(shifts)
    acorr = np.array([np.dot(sequence, np.roll(sequence, ℓ).conj()) for ℓ in shifts])
    energy = np.dot(sequence, sequence.conj())
    return acorr / energy if normalized else acorr



import numpy as np
import numpy.typing as npt

from .. import abc


class UnaryCode(abc.IntegerCode):
    r"""
    Unary code. It is an integer code. For the definition of this code, see [Wikipedia: Unary coding](https://en.wikipedia.org/wiki/Unary_coding).
    """

    def encode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.UnaryCode()
            >>> code.encode([4, 1, 3])
            array([1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 0])
        """
        input = np.asarray(input)
        if input.size == 0:
            return np.array([], dtype=int)
        return np.concatenate([unary_encode(i) for i in input])

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer]:
        r"""
        Examples:
            >>> code = komm.UnaryCode()
            >>> code.decode([1, 1, 1, 1, 0, 1, 0, 1, 1, 1, 0])
            array([4, 1, 3])
        """
        input = np.asarray(input)
        output: list[int] = []
        i = 0
        while i < input.size:
            j = i
            while input[j] == 1:
                j += 1
            output.append(unary_decode(list(input[i : j + 1])))
            i = j + 1
        return np.array(output)


def unary_encode(integer: int) -> list[int]:
    return [1] * integer + [0]


def unary_decode(bits: list[int]) -> int:
    return len(bits) - 1



from dataclasses import dataclass
from functools import cached_property

import numpy as np
import numpy.typing as npt

from .. import abc
from .RaisedCosinePulse import RaisedCosinePulse
from .util import raised_cosine


@dataclass
class RootRaisedCosinePulse(abc.Pulse):
    r"""
    Root-raised-cosine pulse. It is a [pulse](/ref/Pulse) whose spectrum is given by the square root of the spectrum of the [raised cosine pulse](/ref/RaisedCosinePulse) with same roll-off factor.

    The waveform of the root-raised cosine pulse is depicted below for $\alpha = 0.25$, and for $\alpha = 0.75$.

    <div class="centered" markdown>
      <span>
      ![Root-raised-cosine pulse with roll-off factor 0.25.](/fig/pulse_root_raised_cosine_25.svg)
      </span>
      <span>
      ![Root-raised-cosine pulse with roll-off factor 0.75.](/fig/pulse_root_raised_cosine_75.svg)
      </span>
    </div>

    For more details, see [Wikipedia: Root-raised-cosine filter](https://en.wikipedia.org/wiki/Root-raised-cosine_filter).

    Parameters:
        rolloff: The roll-off factor $\alpha$ of the pulse. Must satisfy $0 \leq \alpha \leq 1$. The default value is `1.0`.
    """

    rolloff: float = 1.0

    def waveform(self, t: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the root-raised-cosine pulse, it is given by
        $$
          p(t) = \frac{\sin \( 2 \pi f_1 t \) + 4 \alpha t \cos \( 2 \pi f_2 t \)}{\pi t \( 1 - (4 \alpha t)^2 \)},
        $$
        where $\alpha$ is the roll-off factor, $f_1 = (1 - \alpha) / 2$, and $f_2 = (1 + \alpha) / 2$.

        Examples:
            >>> pulse = komm.RootRaisedCosinePulse(rolloff=0.25)
            >>> pulse.waveform(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ).round(3)
            array([-0.064,  0.238,  0.622,  0.943,  1.068,  0.943,  0.622,  0.238,
                   -0.064])
        """
        α = self.rolloff
        t = np.asarray(t)
        if α == 0:
            return np.sinc(t)
        f1 = (1 - α) / 2
        f2 = (1 + α) / 2
        num = np.sin(2 * np.pi * f1 * t) + 4 * α * t * np.cos(2 * np.pi * f2 * t)
        den = np.pi * t * (1 - (4 * α * t) ** 2)
        with np.errstate(divide="ignore", invalid="ignore"):
            waveform = num / den
        singularity = np.isclose(t, 0)
        waveform[singularity] = 1 + α * (4 / np.pi - 1)
        singularity = np.isclose(np.abs(t), 1 / (4 * α))
        lhs = (1 + 2 / np.pi) * np.sin(np.pi / (4 * α))
        rhs = (1 - 2 / np.pi) * np.cos(np.pi / (4 * α))
        waveform[singularity] = α / np.sqrt(2) * (lhs + rhs)
        return waveform

    def spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.complexfloating]:
        r"""
        Examples:
            >>> pulse = komm.RootRaisedCosinePulse(rolloff=0.25)
            >>> np.abs(pulse.spectrum(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... )).round(3)
            array([0.   , 0.   , 0.707, 1.   , 1.   , 1.   , 0.707, 0.   , 0.   ])
        """
        return np.sqrt(RaisedCosinePulse(self.rolloff).spectrum(f))

    def autocorrelation(self, tau: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the root-raised-cosine pulse, it is given by
        $$
            R(\tau) = \sinc(\tau) \frac{\cos(\pi \alpha \tau)}{1 - (2 \alpha \tau)^2}.
        $$

        Examples:
            >>> pulse = komm.RootRaisedCosinePulse(rolloff=0.25)
            >>> pulse.autocorrelation(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ).round(3)
            array([0.   , 0.29 , 0.627, 0.897, 1.   , 0.897, 0.627, 0.29 , 0.   ])
        """
        return raised_cosine(tau, self.rolloff)

    def energy_density_spectrum(self, f: npt.ArrayLike) -> npt.NDArray[np.floating]:
        r"""
        For the root-raised-cosine pulse, it is given by
        $$
            S(f) = \begin{cases}
                1, & |f| \leq f_1, \\\\[1ex]
                \dfrac{1}{2} \left( 1 + \cos \left( \pi \dfrac{|f| - f_1}{f_2 - f_1}\right) \right), & f_1 \leq |f| \leq f_2, \\\\[1ex]
                0, & \text{otherwise}.
            \end{cases}
        $$

        Examples:
            >>> pulse = komm.RootRaisedCosinePulse(rolloff=0.25)
            >>> pulse.energy_density_spectrum(
            ...     [-1.0, -0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75, 1.0],
            ... ).round(3)
            array([0. , 0. , 0.5, 1. , 1. , 1. , 0.5, 0. , 0. ])
        """
        return np.abs(self.spectrum(f)) ** 2

    @cached_property
    def support(self) -> tuple[float, float]:
        r"""
        For the root-raised-cosine pulse, the support is given by $(-\infty, \infty)$.

        Examples:
            >>> pulse = komm.RootRaisedCosinePulse(rolloff=0.25)
            >>> pulse.support
            (-inf, inf)
        """
        return (-np.inf, np.inf)

    def taps(
        self,
        samples_per_symbol: int,
        span: tuple[int, int] | None = None,
    ) -> npt.NDArray[np.floating]:
        r"""
        Examples:
            >>> pulse = komm.RootRaisedCosinePulse(rolloff=0.25)
            >>> pulse.taps(samples_per_symbol=4, span=(-1, 1)).round(3)
            array([-0.064,  0.238,  0.622,  0.943,  1.068,  0.943,  0.622,  0.238,
                   -0.064])
            >>> pulse.taps(samples_per_symbol=4, span=(-16, 16)).shape
            (129,)
        """
        return super().taps(samples_per_symbol, span)



from ._channels.base import DiscreteMemorylessChannel
from ._constellations.base import Constellation
from ._error_control_block.base import BlockCode
from ._error_control_convolutional.base import ConvolutionalCode
from ._error_control_decoders.base import BlockDecoder
from ._integer_coding.base import IntegerCode
from ._labelings.base import Labeling
from ._pulses.base import Pulse
from ._quantization.base import ScalarQuantizer

__all__ = [
    "BlockCode",
    "BlockDecoder",
    "ConvolutionalCode",
    "DiscreteMemorylessChannel",
    "IntegerCode",
    "Constellation",
    "Labeling",
    "Pulse",
    "ScalarQuantizer",
]



from dataclasses import dataclass
from typing import Literal

import numpy as np
import numpy.typing as npt

from .. import abc
from .._error_control_convolutional.TerminatedConvolutionalCode import (
    TerminatedConvolutionalCode,
)
from .._labelings.NaturalLabeling import NaturalLabeling
from .._util.bit_operations import int_to_bits
from .._util.decorators import blockwise, vectorize, with_pbar
from .util import get_pbar


@dataclass
class BCJRDecoder(abc.BlockDecoder[TerminatedConvolutionalCode]):
    r"""
    Bahl–Cocke–Jelinek–Raviv (BCJR) decoder for [terminated convolutional codes](/ref/TerminatedConvolutionalCode). For more details, see <cite>LC04, Sec. 12.6</cite>.

    Parameters:
        code: The terminated convolutional code to be used for decoding.
        output_type: The type of the output. Either `'hard'` or `'soft'`. Default is `'soft'`.

    Notes:
        - Input type: `soft` (L-values).
        - Output type: `hard` (bits) or `soft` (L-values).
    """

    code: TerminatedConvolutionalCode
    output_type: Literal["hard", "soft"] = "soft"

    def __post_init__(self) -> None:
        if self.code.mode == "tail-biting":
            raise NotImplementedError(
                "BCJR algorithm not implemented for 'tail-biting'"
            )
        if self.output_type not in ["hard", "soft"]:
            raise ValueError("'output_type' must be 'hard' or 'soft'")
        n = self.code.convolutional_code.num_output_bits
        k = self.code.convolutional_code.num_input_bits
        self._fsm = self.code.convolutional_code.finite_state_machine()
        num_states = self._fsm.num_states
        self._initial_state_distribution, self._final_state_distribution = (
            self.code.strategy.bcjr_initial_final_distributions(num_states)
        )
        self._post_process_output = self.code.strategy.bcjr_post_process_output
        self._cache_polar = (-1) ** int_to_bits(range(2**n), width=n)
        self._labeling = NaturalLabeling(k)

    def _metric_function(self, y: int, z: float) -> float:
        return 0.5 * np.dot(self._cache_polar[y], z)

    def decode(self, input: npt.ArrayLike) -> npt.NDArray[np.integer | np.floating]:
        r"""
        Examples:
            >>> code = komm.TerminatedConvolutionalCode(
            ...     convolutional_code=komm.ConvolutionalCode([[0b11, 0b1]], [0b11]),
            ...     num_blocks=3,
            ...     mode="zero-termination",
            ... )

            >>> decoder = komm.BCJRDecoder(code)
            >>> decoder.decode([-0.8, -0.1, -1.0, +0.5, +1.8, -1.1, -1.6, +1.6])
            array([-0.47774884, -0.61545527,  1.03018771])

            >>> decoder = komm.BCJRDecoder(code, output_type="hard")
            >>> decoder.decode([-0.8, -0.1, -1.0, +0.5, +1.8, -1.1, -1.6, +1.6])
            array([1, 1, 0])
        """
        n = self.code.convolutional_code.num_output_bits

        @blockwise(self.code.length)
        @vectorize
        @with_pbar(get_pbar(np.size(input) // self.code.length, "BCJR"))
        def decode(li: npt.NDArray[np.floating]) -> npt.NDArray[np.floating]:
            symbol_posteriors = self._fsm.forward_backward(
                observed=li.reshape(-1, n),
                metric_function=self._metric_function,
                initial_state_distribution=self._initial_state_distribution,
                final_state_distribution=self._final_state_distribution,
            )
            symbol_posteriors = self._post_process_output(symbol_posteriors)
            lo = self._labeling.marginalize(symbol_posteriors).reshape(-1)
            return lo

        output = decode(input)
        if self.output_type == "hard":
            return (output < 0.0).astype(int)
        else:  # if self.output_type == "soft":
            return output
